@@ -45,6 +45,7 @@
 # include <stdlib.h>
 # include <sys/types.h>
 # include <sys/time.h>
+# include <time.h>
 # include <sys/stat.h>
 # include <dirent.h>
 # include <assert.h>
@@ -63,12 +64,12 @@
 /*****************************************************************
 **	private (static) function declaration and definition
 *****************************************************************/
-static	char	dki_estr[255+1];
+static	char	dki_estr[MAX_PATHSIZE+MAX_FNAMESIZE+1];
 
 static	dki_t	*dki_alloc ()
 {
 	dki_estr[0] = '\0';
-	dki_t	*dkp = malloc (sizeof (dki_t));
+	dki_t	*dkp;
 
 	if ( (dkp = malloc (sizeof (dki_t))) )
 	{
@@ -167,9 +168,14 @@ static	int	dki_writeinfo (const dki_t *dkp, const char *path)
 
 	if ( (fp = fopen (path, "w")) == NULL )
 		return 0;
+
 	dbg_val1 ("dki_writeinfo %s\n", path);
 	if ( dki_prt_dnskey_raw (dkp, fp) == 0 )
+	{
+		fclose (fp);
 		return 0;
+	}
+
 	fclose (fp);
 	touch (path, dkp->time);	/* restore time of key file */
 
@@ -227,11 +233,7 @@ void	dki_tfree (dki_t **tree)
 }
 #endif
 
-#if defined(BIND_VERSION) && BIND_VERSION >= 970
-# define	KEYGEN_COMPMODE	"-C -q "	/* this is the compability mode needed by BIND 9.7 */
-#else
-# define	KEYGEN_COMPMODE	""
-#endif
+# define	KEYGEN_COMPMODE	"-C -q "	/* this is the compability mode needed since BIND 9.7 */
 /*****************************************************************
 **	dki_new ()
 **	create new keyfile
@@ -245,30 +247,37 @@ dki_t	*dki_new (const char *dir, const char *name, int ksk, int algo, int bitsiz
 	FILE	*fp;
 	int	len;
 	char	*flag = "";
-	char	*expflag = "";
+	char    *expflag = "";
 	dki_t	*new;
-
+ 
 	if ( ksk )
 		flag = "-f KSK";
 
 	randfile[0] = '\0';
+#if defined(BIND_VERSION) && BIND_VERSION < 91304
 	if ( rfile && *rfile )
 		snprintf (randfile, sizeof (randfile), "-r %.250s ", rfile);
-		
+#endif
+
+#if defined(BIND_VERSION) && BIND_VERSION < 90902
 	if ( algo == DK_ALGO_RSA || algo == DK_ALGO_RSASHA1 || algo == DK_ALGO_RSASHA256 || algo == DK_ALGO_RSASHA512 )
 		expflag = "-e ";
-
+#endif
 	if ( dir && *dir )
-		snprintf (cmdline, sizeof (cmdline), "cd %s ; %s %s%s%s-n ZONE -a %s -b %d %s %s",
+		snprintf (cmdline, sizeof (cmdline), "cd %s ; %s %s%s%s-n ZONE -a %s -b %d %s %s 2>&1",
 			dir, KEYGENCMD, KEYGEN_COMPMODE, randfile, expflag, dki_algo2str(algo), bitsize, flag, name);
 	else
-		snprintf (cmdline, sizeof (cmdline), "%s %s%s%s-n ZONE -a %s -b %d %s %s",
+		snprintf (cmdline, sizeof (cmdline), "%s %s%s%s-n ZONE -a %s -b %d %s %s 2>&1",
 			KEYGENCMD, KEYGEN_COMPMODE, randfile, expflag, dki_algo2str(algo), bitsize, flag, name);
 
 	dbg_msg (cmdline);
 
-	if ( (fp = popen (cmdline, "r")) == NULL || fgets (fname, sizeof fname, fp) == NULL )
+	if ( (fp = popen (cmdline, "r")) == NULL )
 		return NULL;
+
+	fname[0] = '\0';
+	fgets (fname, sizeof fname, fp);
+
 	pclose (fp);
 
 	len = strlen (fname) - 1;
@@ -325,6 +334,7 @@ dki_t	*dki_read (const char *dirname, const char *filename)
 	{
 		snprintf (dki_estr, sizeof (dki_estr),
 			"dki_read: Filename don't match expected format (%s)", fname);
+		dki_free (dkp);
 		return (NULL);
 	}
 
@@ -334,6 +344,7 @@ dki_t	*dki_read (const char *dirname, const char *filename)
 	{
 		snprintf (dki_estr, sizeof (dki_estr),
 			"dki_read: Can\'t open file \"%s\" for reading", path);
+		dki_free (dkp);
 		return (NULL);
 	}
 	
@@ -344,6 +355,7 @@ dki_t	*dki_read (const char *dirname, const char *filename)
 		snprintf (dki_estr, sizeof (dki_estr),
 			"dki_read: Can\'t read key from file %s (errno %d)", path, err);
 		fclose (fp);
+		dki_free (dkp);
 		return (NULL);
 	}
 
@@ -352,6 +364,8 @@ dki_t	*dki_read (const char *dirname, const char *filename)
 	{
 		snprintf (dki_estr, sizeof (dki_estr),
 			"dki_read: Can\'t stat file %s", fname);
+		fclose (fp);
+		dki_free (dkp);
 		return (NULL);
 	}
 	dkp->time = st.st_mtime;
@@ -452,12 +466,10 @@ static	int	dki_setstat (dki_t *dkp, int status, int preserve_time)
 	char	frompath[MAX_PATHSIZE+1];
 	char	topath[MAX_PATHSIZE+1];
 	time_t	totime;
-	time_t	currtime;
 
 	if ( dkp == NULL )
 		return 0;
 
-	currtime = time (NULL);
 	status = tolower (status);
 	switch ( dkp->status )	/* look at old status */
 	{
@@ -639,6 +651,8 @@ char	*dki_algo2str (int algo)
 	case DK_ALGO_NSEC3RSASHA1:	return ("NSEC3RSASHA1");
 	case DK_ALGO_RSASHA256:		return ("RSASHA256");
 	case DK_ALGO_RSASHA512:		return ("RSASHA512");
+	case DK_ALGO_ECDSAP256SHA256:	return ("ECDSAP256SHA256");
+	case DK_ALGO_ECDSAP384SHA384:	return ("ECDSAP384SHA384");
 	}
 	return ("unknown");
 }
@@ -660,6 +674,8 @@ char	*dki_algo2sstr (int algo)
 	case DK_ALGO_NSEC3RSASHA1:	return ("N3RSA1");
 	case DK_ALGO_RSASHA256:		return ("RSASHA2");
 	case DK_ALGO_RSASHA512:		return ("RSASHA5");
+	case DK_ALGO_ECDSAP256SHA256:	return ("P256");
+	case DK_ALGO_ECDSAP384SHA384:	return ("P384");
 	}
 	return ("unknown");
 }
