@@ -45,6 +45,8 @@
 # include <errno.h>	
 # include <unistd.h>	
 # include <ctype.h>	
+# include <sys/types.h>
+# include <time.h>
 
 #ifdef HAVE_CONFIG_H
 # include <config.h>
@@ -57,18 +59,15 @@
 # include "debug.h"
 # include "misc.h"
 # include "ncparse.h"
+# include "dki.h"
+# include "zone.h"
 # include "nscomm.h"
 # include "soaserial.h"
-# include "zone.h"
-# include "dki.h"
 # include "rollover.h"
 # include "log.h"
+# include "zfparse.h"
 
-#if defined(BIND_VERSION) && BIND_VERSION >= 940
 # define	short_options	"c:L:V:D:N:o:O:dfHhnrv"
-#else
-# define	short_options	"c:L:V:D:N:o:O:fHhnrv"
-#endif
 #if defined(HAVE_GETOPT_LONG) && HAVE_GETOPT_LONG
 static struct option long_options[] = {
 	{"reload",		no_argument, NULL, 'r'},
@@ -84,9 +83,7 @@ static struct option long_options[] = {
 	{"directory",		required_argument, NULL, 'D'},
 	{"named-conf",		required_argument, NULL, 'N'},
 	{"origin",		required_argument, NULL, 'o'},
-#if defined(BIND_VERSION) && BIND_VERSION >= 940
 	{"dynamic",		no_argument, NULL, 'd' },
-#endif
 	{"help",		no_argument, NULL, 'h'},
 	{0, 0, 0, 0}
 };
@@ -125,9 +122,7 @@ static	zone_t	*zonelist = NULL;	/* must be static global because add2zonelist us
 static	zconf_t	*config;
 
 /**	macros **/
-#define	set_bind94_dynzone(dz)	((dz) = 1)
 #define	set_bind96_dynzone(dz)	((dz) = 6)
-#define	bind94_dynzone(dz)	( (dz) > 0 && (dz) < 6 )
 #define	bind96_dynzone(dz)	( (dz) >= 6 )
 #define	is_defined(str)		( (str) && *(str) )
 
@@ -208,17 +203,11 @@ int	main (int argc, char *const argv[])
 		case 'h':
 			usage (NULL, config);
 			break;
-#if defined(BIND_VERSION) && BIND_VERSION >= 940
 		case 'd':
-# if BIND_VERSION >= 960
-			set_bind96_dynzone (dynamic_zone);
-# else
-			set_bind94_dynzone(dynamic_zone);
-# endif
+			dynamic_zone = 1;
 			/* dynamic zone requires a name server reload... */
 			reloadflag = 0;		/* ...but "rndc thaw" reloads the zone anyway */
 			break;
-#endif
 		case 'n':
 			noexec = 1;
 			break;
@@ -492,7 +481,17 @@ static	int	dosigning (zone_t *zonelist, zone_t *zp)
 		return 2;
 	}
 	
+#if defined (USE_INCLUDE_FILE_TRACKING) && USE_INCLUDE_FILE_TRACKING
+	zfile_time = recursive_file_mtime (zp->dir, zp->file, zp->conf->keyfile);
+#else
 	zfile_time = file_mtime (path);
+#endif
+	if ( !zfile_time )
+	{
+		error ("Error parsing zone file (%s)!\n", path);
+		lg_mesg (LG_ERROR, "\"%s\": error parsing zone file (%s)!", zp->zone, path);
+		return 2;
+	}
 	currtime = time (NULL);
 
 	/* check for domain based logging */
@@ -538,13 +537,40 @@ static	int	dosigning (zone_t *zonelist, zone_t *zp)
 		newkeysetfile = new_keysetfiles (zp->dir, zfilesig_time);
 #endif
 
+	/* is there a list of files included in zone.db ? */
+	if ( zp->conf->dependfiles && *zp->conf->dependfiles )
+	{
+		char	file[255+1];
+		const	char	*p;
+		int	i;
+		time_t	incfile_mtime;
+
+		/* check the timestamp of each file against "zone.db" */
+		p = zp->conf->dependfiles;
+		while ( p && *p )
+		{
+			while ( isflistdelim (*p) )
+				p++;
+
+			for ( i = 0; i < 255 && *p && !isflistdelim (*p); i++ )
+				file[i] = *p++;
+			file[i] = '\0';
+
+			pathname (path, sizeof (path), zp->dir, file, NULL);
+
+			incfile_mtime = file_mtime (path);
+			if ( incfile_mtime > zfile_time )	/* include file is newer? */
+				zfile_time = incfile_mtime;	/* take this one as new mtime */
+		}
+	}
+
 	/**
 	** Check if it is time to do a re-sign. This is the case if
 	**	a) the command line flag -f is set, or
 	**	b) new keys are generated, or
 	**	c) we found a new KSK of a delegated domain, or
 	**	d) the "dnskey.db" file is newer than "zone.db" 
-	**	e) the "zone.db" is newer than "zone.db.signed" or
+	**	e) the "zone.db" or any included file is newer than "zone.db.signed" or
 	**	f) "zone.db.signed" is older than the re-sign interval
 	**/
 	mesg[0] = '\0';
@@ -561,8 +587,6 @@ static	int	dosigning (zone_t *zonelist, zone_t *zp)
 	else if ( (currtime - zfilesig_time) > zp->conf->resign - (OFFSET) )
 		snprintf (mesg, sizeof(mesg), "re-signing interval (%s) reached",
 						str_delspace (age2str (zp->conf->resign)));
-	else if ( bind94_dynzone (dynamic_zone) )
-		snprintf (mesg, sizeof(mesg), "dynamic zone");
 
 	if ( *mesg )
 		verbmesg (1, zp->conf, "\tRe-signing necessary: %s\n", mesg);
@@ -575,8 +599,7 @@ static	int	dosigning (zone_t *zonelist, zone_t *zp)
 	dbg_line ();
 	if ( !(force || newkey || newkeysetfile || zfile_time > zfilesig_time ||	
 	     file_mtime (path) > zfilesig_time ||
-	     (currtime - zfilesig_time) > zp->conf->resign - (OFFSET) ||
-	      bind94_dynzone (dynamic_zone)) )
+	     (currtime - zfilesig_time) > zp->conf->resign - (OFFSET)) )
 	{
 		verbmesg (2, zp->conf, "\tCheck if there is a parent file to copy\n");
 		if ( zp->conf->keysetdir && strcmp (zp->conf->keysetdir, "..") == 0 )
@@ -601,11 +624,7 @@ static	int	dosigning (zone_t *zonelist, zone_t *zp)
 	err = 1;
 	use_unixtime = ( zp->conf->serialform == Unixtime );
 	dbg_val1 ("Use unixtime = %d\n", use_unixtime);
-#if defined(BIND_VERSION) && BIND_VERSION >= 940
 	if ( !dynamic_zone && !use_unixtime ) /* increment serial number in static zone files */
-#else
-	if ( !dynamic_zone ) /* increment serial no in static zone files */
-#endif
 	{
 		pathname (path, sizeof (path), zp->dir, zp->file, NULL);
 		err = 0;
@@ -627,7 +646,7 @@ static	int	dosigning (zone_t *zonelist, zone_t *zp)
 	}
 
 	/* at last, sign the zone file */
-	if ( err > 0 )
+	if ( err >= 0 )
 	{
 		time_t	timer;
 
@@ -710,11 +729,20 @@ static	int	dosigning (zone_t *zonelist, zone_t *zp)
 	return err;
 }
 
+/*****************************************************************
+**	This function is no longer needed, and us doing in fact
+**	nothing.
+**	The primary idea was to register a new KSK at a experimental
+**	key registration instance.
+**	This code is still here, to have an anchor for a (new) client
+**	parent DS update protocol (e.g. via dynamic updates).
+*****************************************************************/
 static	void	register_key (dki_t *list, const zconf_t *z)
 {
+#if 0
 	dki_t	*dkp;
-	time_t	currtime;
 	time_t	age;
+	time_t	currtime;
 
 	assert ( list != NULL );
 	assert ( z != NULL );
@@ -723,7 +751,7 @@ static	void	register_key (dki_t *list, const zconf_t *z)
 	for ( dkp = list; dkp && dki_isksk (dkp); dkp = dkp->next )
 	{
 		age = dki_age (dkp, currtime);
-#if 0
+
 		/* announce "new" and active key signing keys */
 		if ( REG_URL && *REG_URL && dki_status (dkp) == DKI_ACT && age <= z->resign * 4 )
 		{
@@ -731,8 +759,8 @@ static	void	register_key (dki_t *list, const zconf_t *z)
 				logmesg ("\tRegister new KSK with tag %d for domain %s\n",
 								dkp->tag, dkp->name);
 		}
-#endif
 	}
+#endif
 }
 
 /*
@@ -826,7 +854,7 @@ static	int	writekeyfile (const char *fname, const dki_t *list, int key_ttl)
 static	int	sign_zone (const zone_t *zp)
 {
 	char	cmd[2047+1];
-	char	str[1023+1];
+	char	str[254+1];
 	char	rparam[254+1];
 	char	nsec3param[637+1];
 	char	keysetdir[254+1];
@@ -835,7 +863,6 @@ static	int	sign_zone (const zone_t *zp)
 	const	char	*pseudo;
 	const	char	*param;
 	int	len;
-	FILE	*fp;
 
 	const	char	*dir;
 	const	char	*domain;
@@ -854,24 +881,16 @@ static	int	sign_zone (const zone_t *zp)
 		len = snprintf (str, sizeof (str), "-l %.250s", conf->lookaside);
 
 	dbg_line();
-#if defined(BIND_VERSION) && BIND_VERSION >= 940
 	if ( !dynamic_zone && conf->serialform == Unixtime )
 		snprintf (str+len, sizeof (str) - len, " -N unixtime");
-#endif
 
 	gends = "";
 	if ( conf->sig_gends )
-#if defined(BIND_VERSION) && BIND_VERSION >= 970
 		gends = "-C -g ";
-#else
-		gends = "-g ";
-#endif
 
 	dnskeyksk = "";
-#if defined(BIND_VERSION) && BIND_VERSION >= 970
 	if ( conf->sig_dnskeyksk )
 		dnskeyksk = "-x ";
-#endif
 
 	pseudo = "";
 	if ( conf->sig_pseudo )
@@ -882,7 +901,6 @@ static	int	sign_zone (const zone_t *zp)
 		param = conf->sig_param;
 
 	nsec3param[0] = '\0';
-#if defined(BIND_VERSION) && BIND_VERSION >= 960
 	if ( conf->k_algo == DK_ALGO_NSEC3DSA || conf->k_algo == DK_ALGO_NSEC3RSASHA1 ||
 	     conf->nsec3 != NSEC3_OFF )
 	{
@@ -891,11 +909,7 @@ static	int	sign_zone (const zone_t *zp)
 		const	char	*optout;
 		unsigned int	seed;
 
-# if defined(BIND_VERSION) && BIND_VERSION >= 970
 		update = "-u ";		/* trailing blank is necessary */
-# else
-		update = "";
-# endif
 		if ( conf->nsec3 == NSEC3_OPTOUT )
 			optout = "-A ";
 		else
@@ -919,7 +933,6 @@ static	int	sign_zone (const zone_t *zp)
 		if ( gensalt (salt, sizeof (salt), conf->saltbits, seed) )
 			snprintf (nsec3param, sizeof (nsec3param), "%s%s-3 %s ", update, optout, salt);
 	}
-#endif
 
 	dbg_line();
 	rparam[0] = '\0';
@@ -935,33 +948,31 @@ static	int	sign_zone (const zone_t *zp)
 		dir = ".";
 
 	dbg_line();
-#if defined(BIND_VERSION) && BIND_VERSION >= 940
 	if ( dynamic_zone )
 		snprintf (cmd, sizeof (cmd), "cd %s; %s %s %s%s%s%s%s%s-o %s -e +%ld %s -N increment -f %s.dsigned %s K*.private 2>&1",
 			dir, SIGNCMD, param, nsec3param, dnskeyksk, gends, pseudo, rparam, keysetdir, domain, conf->sigvalidity, str, file, file);
 	else
-#endif
 		snprintf (cmd, sizeof (cmd), "cd %s; %s %s %s%s%s%s%s%s-o %s -e +%ld %s %s K*.private 2>&1",
 			dir, SIGNCMD, param, nsec3param, dnskeyksk, gends, pseudo, rparam, keysetdir, domain, conf->sigvalidity, str, file);
 	verbmesg (2, conf, "\t  Run cmd \"%s\"\n", cmd);
 	*str = '\0';
 	if ( noexec == 0 )
 	{
-#if 0
-		if ( (fp = popen (cmd, "r")) == NULL || fgets (str, sizeof str, fp) == NULL )
-			return -1;
-#else
+		FILE	*fp;
+		int	exitcode;
+
 		if ( (fp = popen (cmd, "r")) == NULL )
 			return -1;
+
 		str[0] = '\0';
 		while ( fgets (str, sizeof str, fp) != NULL )	/* eat up all output until the last line */
 			;
-#endif
-		pclose (fp);
+
+		exitcode = pclose (fp);
+		verbmesg (2, conf, "\t  Cmd dnssec-signzone returns with exitcode=%d: \"%s\"\n", exitcode, str_chop (str, '\n'));
 	}
 
 	dbg_line();
-	verbmesg (2, conf, "\t  Cmd dnssec-signzone return: \"%s\"\n", str_chop (str, '\n'));
 	len = strlen (str) - 6;
 	if ( len < 0 || strcmp (str+len, "signed") != 0 )
 		return -1;
